"""Top-level package for plotski."""
from plotski.base import Plot  # noqa
from plotski.store import PlotStore  # noqa
from plotski.store.custom import CustomPlotStore  # noqa

__author__ = "Lukasz G. Migas"
__email__ = "l.g.migas@tudelft.nl"
__version__ = "0.1.0"
